package main.java.exercise;

import main.java.framework.StudentInformation;
import main.java.framework.StudentSolution;

public class StudentSolutionImplementation implements StudentSolution {
    @Override
    public StudentInformation provideStudentInformation() {
        return new StudentInformation(
                "Vorname", // Vorname
                "Nachname", // Nachname
                "1234567" // Matrikelnummer
        );
    }

    // Implementieren Sie hier Ihre Lösung mit Polynomialzeitreduktion
    public int findMaxClique(Graph g, TimedSolver solver, boolean[] chosenVertices) {
        // If graph is empty return 0
        if (g.numberOfVertices() < 1) {
            return 0;
        }

        String biggestClique = "";
        int[] bounds = {1, g.numberOfVertices()};
        assert bounds[0] <= bounds[1];

        while (bounds[0] != bounds[1]) {
            int k = (bounds[0] + bounds[1] + 1)/2;  // The +1 converts the floor division to a ceil division

            /*// Do not need to run getCliqueOfSize for k=1
            if (k < 2) {
                break;
            }*/

            String result = getCliqueOfSize(g, k, solver);
            assert result != null;

            // Is there a clique of size k or smaller?
            if (result.isEmpty()) {
                // g has no clique of size k or smaller, update upper bound
                bounds[1] = k-1;
            } else {
                // g has a clique of size k, update lower bound
                bounds[0] = k;

                // The last found clique is always the biggest one
                assert result.length() > biggestClique.length();
                biggestClique = result;
            }
        }
        //assert bounds[0] == bounds[1];
        assert (biggestClique.isEmpty()) == (bounds[0] == 1);  // biggestClique.isEmpty() <=> maxCliqueSize==1

        findChosenVertices(biggestClique, chosenVertices);

        return bounds[0];
    }


    private String getCliqueOfSize(Graph g, int size, TimedSolver solver) {
        String formula = buildCNF(g, size);
        return solver.solve(formula);
    }

    private String buildCNF(Graph g, int size) {
        int vertexCount = g.numberOfVertices();

        StringBuilder formula = new StringBuilder("p cnf ");

        int clauseCount = size +
                size*(size-1)/2 * vertexCount +
                size*(size-1) * ( vertexCount*(vertexCount-1)/2 - g.numberOfEdges() );

        // Add p cnf n_var n_claus header
        formula.append(size * vertexCount);
        formula.append(' ');
        formula.append(clauseCount);
        formula.append('\n');

        // For each i, there is an ith vertex in the clique
        for (int i = 0; i < size; ++i) {
            // The clauses for the i'th vertices are: i*vertexCount to (i+1)*vertexCount
            for (int v = 1; v <= vertexCount; ++v) {
                formula.append(i*vertexCount + v);
                formula.append(" ");
            }
            formula.append("0\n");
        }

        //clauseCount = size;

        // For each i,j, the ith vertex is different from the jth vertex
        for (int i = 0; i < size; ++i) {
            for (int j = i+1; j < size; ++j) {
                for (int v = 1; v <= vertexCount; ++v) {
                    int x_iv = i*vertexCount + v;
                    int x_jv = j*vertexCount + v;
                    formula.append('-');
                    formula.append(x_iv);
                    formula.append(" -");
                    formula.append(x_jv);
                    formula.append(" 0\n");
                }
            }
            //clauseCount += vertexCount * (size - i - 1);
        }
        //clauseCount += vertexCount * (size*(size-1) - (size-1)*size/2);
        //clauseCount += vertexCount * ((size-1)*size/2);

        // For each non-edge (u,v), u,v cannot both belong to the clique
        for (int u = 1; u <= vertexCount; ++u) {
            for (int v = u+1; v <= vertexCount; ++v) {
                if (!g.containsEdge(u,v)) {
                    for (int i = 0; i < size; ++i) {
                        for (int j = 0; j < size; ++j) {
                            if (i != j) {
                                int x_iu = i*vertexCount + u;
                                int x_jv = j*vertexCount + v;
                                formula.append('-');
                                formula.append(x_iu);
                                formula.append(" -");
                                formula.append(x_jv);
                                formula.append(" 0\n");
                            }
                        }
                    }
                    //clauseCount += size*(size-1);
                }
            }
        }
        //clauseCount += size*(size-1) * ( (vertexCount-1)*vertexCount/2 - g.numberOfEdges() );

        return formula.toString();
    }

    private void findChosenVertices(String result, boolean[] chosenVertices) {
        int vertexCount = chosenVertices.length;

        if (result.isEmpty()) {
            chosenVertices[0] = true;
        } else {
            String[] parts = result.split(" ");
            for (int i = 0; i < parts.length-1; ++i) {
                // Vertex i is chosen if it's variable is not negated
                if (parts[i].charAt(0) != '-') {
                    chosenVertices[i % vertexCount] = true;
                }
            }
        }
    }
}
